program WebFrameCapture;

{
Revision history:

V1.0.0  1997 Jun 07  First version
V1.0.2  1997 Jul 04  Delete intermediate files after use
V1.0.4  1997 Oct 18  Allow parameter on URL of SMALL, MEDIUM or LARGE
V1.0.6  1997 Nov 29  Add optional date and time stamp on the image,
                     Enabled by TIMESTAMP or BOLDTIMESTAMP in the parameters
                     Add functional description notes
                     Re-organise folders into separate source folder
V1.1.0  2000 Aug 07  Version for Delphi 4.03
                     Recognise need for choice of video source

Purpose:

To provide a CGI interface allowing a live WebCam picture to be obtained from
an installed VideoForWindows device - a Connectix Color QuickCam in my case.
This program compiles to a Win32 "console" application, like an old DOS
program, that uses a handle to the desktop Window to capture a frame of video.
This video is then processed into a JPEG file returned to the browser.  Since
the program is executed, it must be placed into a Web server folder that allows
execution of files.  This would be the /Scripts folder on Microsoft's IIS.
Note that there is no attempt to allow multiple, simultaneous callers.

The image URL would then be something like:

  <img  src="/Scripts/WebFrameCapture.exe?"
  alt="[please while live image loads]" border="4">

URL Parameters:

You can use parameters on the URL to control the execution of the program.
In particular, you can change the size of the image that is returned, and you
can cause a timestamp to be added to the iamge.  Without parameters, an image
sized 160 x 120 with no timestamp is returned.

  src="/Scripts/WebFrameCapture.exe?medium"  - 320 x 240 image size
  src="/Scripts/WebFrameCapture.exe?large"   - 640 x 480 image size

  src="/Scripts/WebFrameCapture.exe?timestamp"      - add timestamp
  src="/Scripts/WebFrameCapture.exe?boldtimestamp"  - add BOLD timestamp

 sample URL using both sets of parameters :-

  <img src="/Scripts/WebFrameCapture.exe?medium?timestamp"
  alt="[please while live image loads]" border="4">

Notes:

Compression quality:

JPEG allows the compression to be set according to the quality desired, ranging
from very blocky and unreconisable (quality=1 in Borland's JPEG unit) to perfect
(quality=100).  Higher numbers give better quality but a larger file size.  You
may wish to alter the value contained in this program, designed for a Color
QuickCam where the value of 85 seems to keep the picture source, rather than the
JPEG channel, as the limiting factor in picture quality.

The Timestamp image:

Try to add an image that is clear over a high-compression JPEG channel, whilst
being not overly distracting to the picture.  That is achieved by writing an
off-white image on an off-black background.  The font can, optionally, be bold.

Units:

The only non-Borland unit used is VideoCapture, a limited translation of
MS's headers as contained in VfW.h (Video for Windows).

Video Drivers:

Note that if you have more than one video device on your system, you may have
to define the constant driver_number to be other than zero to obtain the
required video capture device.
}

uses Windows, SysUtils, Graphics, JPEG, VideoCapture;

{$R *.RES}

const
  driver_number = 1;               // zero if only one driver

  dib_name = 'c:\~del-me.bmp';     // temporary file name for bitmap image
  jpeg_name = 'c:\~del-me.jpg';    // temporary file name for JPEG image

  // Header for stream returned to browser
  content_type = 'Content-type: image/jpg'#10#10;

  // Default WebCam picture size
  pic_width: integer = 160;
  pic_height: integer = 120;

  // We choose to use Netscape palette colours here for the timestamp text
  dark_font_colour = TColor ($333333);
  light_font_colour = TColor ($CCCCCC);

  // Empirical amount of compression for acceptable quality
  quality = 85;

var
  dib_filename: string;
  Handle: THandle;
  hWndC: THandle;
  nID: integer;
  std_out: file;
  jpeg_file: file;
  bfr: array [0..4095] of byte;
  bytes_read: integer;
  pic: TPicture;
  jpg: TJPEGImage;
  bmh: TBitmapInfoHeader;
  query_string: array [0..50] of Char;
  show_timestamp: boolean;
  bold_timestamp: boolean;
  time_stamp: string;
  i, j: integer;
  x, y: integer;

begin
  // Grab an image, using the desktop window handle
  Handle := GetDesktopWindow;
  nID := 126;

  // See what values have been set in the enviroment,
  // to extract size and timestamp settings
  GetEnvironmentVariable ('QUERY_STRING', query_string, SizeOf (query_string) - 1);
  StrUpper (query_string);

  // Interpret the timestamp strings
  show_timestamp := StrPos (query_string, 'TIMESTAMP') <> nil;
  bold_timestamp := StrPos (query_string, 'BOLDTIMESTAMP') <> nil;

  // Interpret the image size strings
  if StrPos (query_string, 'MEDIUM') <> nil
  then
    begin
    pic_width := 320;
    pic_height := 240;
    end
  else if StrPos (query_string, 'LARGE') <> nil
       then
         begin
         pic_width := 640;
         pic_height := 480;
         end;

  // Prepare our name for temporary bitmap output file
  dib_filename := dib_name;

  // Try to create a capture window
  hWndC := capCreateCaptureWindow ( 'My Own Capture Window',
    WS_CHILD or WS_VISIBLE , 0, 0, pic_width, pic_height, Handle, nID);
  if hWndC <> 0 then
    begin
    // Try to connect to the driver.  Note that if you have more than one
    // driver, and wish to connect to other than the default driver,
    // the driver_number may need to be greater than zero.
    if SendMessage (hWndC, WM_CAP_DRIVER_CONNECT, driver_number, 0) <> 0 then
      begin
      with bmh do     // Fill the fields of a bitmap header
        begin
        FillChar (bmh, SizeOf (bmh), 0);
        biSize := SizeOf (bmh);
        biWidth := pic_width;
        biHeight := pic_height;
        biPlanes := 1;
        biBitCount := 16;            // You may not wish to use this format
        biCompression := BI_RGB;
        end;
      // Tell the video capture device what image format we want
      SendMessage (hWndC, WM_CAP_SET_VIDEOFORMAT, SizeOf (bmh), integer (@bmh));
      // Now we actually try to grab a frame of video
      if (SendMessage (hWndC, WM_CAP_GRAB_FRAME, 0, 0) <> 0) then
        begin
        // If we got an image, save the image to a temporary file
        if (SendMessage (hWndC, WM_CAP_FILE_SAVEDIB, 0, integer (dib_filename)) <> 0) then
          begin
          // Convert grabbed image from a BMP format into a JPEG format
          pic := TPicture.Create;
          jpg := TJPEGImage.Create;
          pic.LoadFromFile (dib_name);
          if show_timestamp then                // Add a timestamp to the bitmap
            with pic.Bitmap.Canvas do
              begin
              Brush.Style := bsClear;           // For a transparent background
              // Format the text string to write
              time_stamp := FormatDateTime (' yyyy-mmm-dd  hh:mm ', Now);
              // If required, merge in the bold style
              if bold_timestamp then Font.Style := [fsBold];
              Font.Color := dark_font_colour;
              x := 1;
              y := pic_height - TextHeight ('Cy') - 1;
              // Write the text in the dark colour, blurred over
              // three pixels horizontally and vertically.
              for i := -1 to 1 do
                for j := -1 to 1 do
                  TextOut (x + i, y + j, time_stamp);
              // Now write the text in the light colour, in the
              // middle of the dark area we've just prepared.
              Font.Color := light_font_colour;
              TextOut (x, y, time_stamp);
              end;
          jpg.Assign (pic.Bitmap);         // Load the picture into the JPEG
          pic.Free;
          jpg.CompressionQuality := quality;
          jpg.SaveToFile (jpeg_name);     // Save the JPEG to a temporary file
          jpg.Free;

          // Read that JPEG in binary and return it to standard output
          AssignFile (std_out, '');
          Rewrite (std_out, 1);          // binary with a record length of 1
          // Write the content header (i.e. that it's a JPEG stream)
          BlockWrite (std_out, content_type, Length (content_type));
          AssignFile (jpeg_file, jpeg_name);
          Reset (jpeg_file, 1);
          // Copy the JPEG file to standard output
          repeat
            BlockRead (jpeg_file, bfr, SizeOf (bfr), bytes_read);
            if bytes_read <> 0 then BlockWrite (std_out, bfr, bytes_read);
          until bytes_read = 0;
          CloseFile (jpeg_file);
          CloseFile (std_out);
          end;
        end;
      // Release the video capture device
      SendMessage (hWndC, WM_CAP_DRIVER_DISCONNECT, 0, 0);
      end;
    end;

    // Done with the .BMP file - so delete it
    if FileExists (dib_name) then DeleteFile (dib_name);
    // Done with the .JPG file - so delete it
    if FileExists (jpeg_name) then DeleteFile (jpeg_name);
end.

